/**
* @file model_process.cpp
*
* Copyright (C) 2020. Huawei Technologies Co., Ltd. All rights reserved.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/
#include "model_process.h"
#include <iostream>
#include <map>
#include <sstream>
#include <algorithm>
#include <functional>
#include <fstream>
#include <cmath>
#include "utils.h"

using namespace std;
extern bool g_isDevice;

std::vector<float*> outputDataVector;

ModelProcess::ModelProcess() :modelId_(0), modelWorkSize_(0), modelWeightSize_(0), modelWorkPtr_(nullptr),
    modelWeightPtr_(nullptr), loadFlag_(false), modelDesc_(nullptr), input_(nullptr), output_(nullptr)
{
}

ModelProcess::~ModelProcess()
{
    UnloadModel();
    DestroyModelDesc();
    DestroyInput();
    DestroyOutput();
}


Result ModelProcess::read_point(const std::string &file_path, const int num_point_dim,
                void **buffer, int *num_points) {
  std::ifstream file_in(file_path, std::ios::in | std::ios::binary);

  if (!file_in) {
    ERROR_LOG("Failed to read file: %s", file_path.c_str());
    return FAILED;
  }

  std::streampos file_size;
  file_in.seekg(0, std::ios::end);
  file_size = file_in.tellg();
  file_in.seekg(0, std::ios::beg);
  *buffer = malloc(file_size);

  if (*buffer == nullptr) {
    ERROR_LOG("Failed to malloc memory of size: %ld", static_cast<long>(file_size));
    return FAILED;
  }
  file_in.read(reinterpret_cast<char *>(*buffer), file_size);
  file_in.close();

  if (file_size / sizeof(float) % num_point_dim != 0) {
    ERROR_LOG("Loaded file size (%ld) is not evenly divisible by num_point_dim (%d)",
             static_cast<long>(file_size),num_point_dim);
    return FAILED;
  }
  *num_points = file_size / sizeof(float) / num_point_dim;
  return SUCCESS;
}

Result ModelProcess::LoadModel(const char *modelPath)
{
    if (loadFlag_) {
        ERROR_LOG("model has already been loaded");
        return FAILED;
    }
    aclError ret = aclmdlQuerySize(modelPath, &modelWorkSize_, &modelWeightSize_);
    if (!Utils::validStatus("aclmdlQuerySize", ret)) {
        ERROR_LOG("Model file is %s", modelPath);
        return FAILED;
    }
    // using ACL_MEM_MALLOC_HUGE_FIRST to malloc memory, huge memory is preferred to use
    // and huge memory can improve performance.
    ret = aclrtMalloc(&modelWorkPtr_, modelWorkSize_, ACL_MEM_MALLOC_HUGE_FIRST);
    if (!Utils::validStatus("aclrtMalloc", ret)) {
        ERROR_LOG("Require size for work buffer is %zu", modelWorkSize_);
        return FAILED;
    }

    // using ACL_MEM_MALLOC_HUGE_FIRST to malloc memory, huge memory is preferred to use
    // and huge memory can improve performance.
    ret = aclrtMalloc(&modelWeightPtr_, modelWeightSize_, ACL_MEM_MALLOC_HUGE_FIRST);
    if (!Utils::validStatus("aclrtMalloc", ret)) {
        ERROR_LOG("Require size for weight buffer is %zu", modelWeightSize_);
        return FAILED;
    }

    ret = aclmdlLoadFromFileWithMem(modelPath, &modelId_, modelWorkPtr_,
        modelWorkSize_, modelWeightPtr_, modelWeightSize_);
    if (!Utils::validStatus("aclmdlLoadFromFileWithMem", ret)) {
        ERROR_LOG("Model file is %s", modelPath);
        return FAILED;
    }

    loadFlag_ = true;
    INFO_LOG("load model %s success", modelPath);
    return SUCCESS;
}

Result ModelProcess::CreateModelDesc()
{
    modelDesc_ = aclmdlCreateDesc();
    if (modelDesc_ == nullptr) {
        ERROR_LOG("create model description failed");
        return FAILED;
    }

    aclError ret = aclmdlGetDesc(modelDesc_, modelId_);
    if (!Utils::validStatus("aclmdlGetDesc", ret)) {
        ERROR_LOG("modelId is %u", modelId_);
        return FAILED;
    }

    INFO_LOG("create model description success");

    return SUCCESS;
}

void ModelProcess::DestroyModelDesc()
{
    if (modelDesc_ != nullptr) {
        (void)aclmdlDestroyDesc(modelDesc_);
        modelDesc_ = nullptr;
    }
    INFO_LOG("destroy model description success");
}


Result ModelProcess::hard_voxelize(const float point_cloud_range_x_min,
                                 const float point_cloud_range_y_min,
                                 const float point_cloud_range_z_min,
                                 const float voxel_size_x, const float voxel_size_y,
                                 const float voxel_size_z, const int grid_size_x,
                                 const int grid_size_y, const int grid_size_z,
                                 const int max_num_points_in_voxel, const int max_voxels,
                                 const float* points, const int num_point_dim,
                                 const int num_points, float* voxels, int* coords,
                                 int* num_points_per_voxel, int* voxel_num)
{
    voxel_num[0] = 0;
    int voxel_idx, grid_idx, curr_num_point;
    int coord_x, coord_y, coord_z;
    int* grid_idx_to_voxel_idx = new int[grid_size_x * grid_size_y * grid_size_z];

    memset(grid_idx_to_voxel_idx, -1,
           grid_size_x * grid_size_y * grid_size_z * sizeof(int));

    for (int point_idx = 0; point_idx < num_points; ++point_idx) {
        coord_x = floor(
            (points[point_idx * num_point_dim + 0] - point_cloud_range_x_min) /
            voxel_size_x);
        coord_y = floor(
            (points[point_idx * num_point_dim + 1] - point_cloud_range_y_min) /
            voxel_size_y);
        coord_z = floor(
            (points[point_idx * num_point_dim + 2] - point_cloud_range_z_min) /
            voxel_size_z);

        if (coord_x < 0 || coord_x > grid_size_x || coord_x == grid_size_x) {
            continue;
        }
        if (coord_y < 0 || coord_y > grid_size_y || coord_y == grid_size_y) {
            continue;
        }
        if (coord_z < 0 || coord_z > grid_size_z || coord_z == grid_size_z) {
            continue;
        }

        grid_idx =
            coord_z * grid_size_y * grid_size_x + coord_y * grid_size_x + coord_x;
        voxel_idx = grid_idx_to_voxel_idx[grid_idx];
        if (voxel_idx == -1) {
            voxel_idx = voxel_num[0];
            if (voxel_num[0] == max_voxels || voxel_num[0] > max_voxels) {
                continue;
            }
            voxel_num[0]++;
            grid_idx_to_voxel_idx[grid_idx] = voxel_idx;
            coords[voxel_idx * 3 + 0] = coord_z;
            coords[voxel_idx * 3 + 1] = coord_y;
            coords[voxel_idx * 3 + 2] = coord_x;
        }
        curr_num_point = num_points_per_voxel[voxel_idx];
        if (curr_num_point < max_num_points_in_voxel) {
            for (int j = 0; j < num_point_dim; ++j) {
                voxels[voxel_idx * max_num_points_in_voxel * num_point_dim +
                    curr_num_point * num_point_dim + j] =
                    points[point_idx * num_point_dim + j];
            }
            num_points_per_voxel[voxel_idx] = curr_num_point + 1;
        }
    }
    delete[] grid_idx_to_voxel_idx;

    return SUCCESS;
}




Result ModelProcess::GetInputSizeByIndex(const size_t index, size_t &inputSize)
{
    if (modelDesc_ == nullptr) {
        ERROR_LOG("no model description, create input failed");
        return FAILED;
    }
    inputSize = aclmdlGetInputSizeByIndex(modelDesc_, index);
    return SUCCESS;
}


Result ModelProcess::AddDatasetBuffer(aclmdlDataset *dataset, 
                                        void* buffer, uint32_t bufferSize) {
    aclDataBuffer* dataBuf = aclCreateDataBuffer(buffer, bufferSize);
    // if (dataBuf == nullptr) {
    //     ATLAS_LOG_ERROR("Create data buffer error");
    //     return ATLAS_ERROR_CREATE_DATA_BUFFER;
    // }

    aclError ret = aclmdlAddDatasetBuffer(dataset, dataBuf);
    // if (ret != ACL_ERROR_NONE) {
    //     ATLAS_LOG_ERROR("Add dataset buffer error %d", ret);
    //     aclDestroyDataBuffer(dataBuf);
    //     return ATLAS_ERROR_ADD_DATASET_BUFFER;
    // }

    return SUCCESS;
}


Result ModelProcess::CreateInput(vector<DataInfo>& inputData)
{
    uint32_t dataNum = inputData.size();
    // om used in this sample has only one input
    if (modelDesc_ == nullptr) {
        ERROR_LOG("no model description, create input failed");
        return FAILED;
    }
    size_t modelInputSize = aclmdlGetInputSizeByIndex(modelDesc_, 0);
    // if (bufferSize != modelInputSize) {
    //     ERROR_LOG("input image size[%zu] is not equal to model input size[%zu]", bufferSize, modelInputSize);
    //     return FAILED;
    // }

    input_ = aclmdlCreateDataset();
    
    if (input_ == nullptr) {
        ERROR_LOG("can't create dataset, create input failed");
        return FAILED;
    }




    for (uint32_t i = 0; i < inputData.size(); i++) {
        aclError atlRet = AddDatasetBuffer(input_, 
                                             inputData[i].data,
                                             inputData[i].size);
    }



    // aclDataBuffer *inputData = aclCreateDataBuffer(inputDataBuffer, bufferSize);
    // if (inputData == nullptr) {
    //     ERROR_LOG("can't create data buffer, create input failed");
    //     return FAILED;
    // }

    // aclError ret = aclmdlAddDatasetBuffer(input_, inputData);
    // if (!Utils::validStatus("aclmdlAddDatasetBuffer", ret)) {
    //     (void)aclDestroyDataBuffer(inputData);
    //     inputData = nullptr;
    //     return FAILED;
    // }
    // INFO_LOG("create model input success");

    return SUCCESS;
}

void ModelProcess::DestroyInput()
{
    if (input_ == nullptr) {
        return;
    }

    for (size_t i = 0; i < aclmdlGetDatasetNumBuffers(input_); ++i) {
        aclDataBuffer *dataBuffer = aclmdlGetDatasetBuffer(input_, i);
        (void)aclDestroyDataBuffer(dataBuffer);
    }
    (void)aclmdlDestroyDataset(input_);
    input_ = nullptr;
    INFO_LOG("destroy model input success");
}

Result ModelProcess::CreateOutput()
{
    if (modelDesc_ == nullptr) {
        ERROR_LOG("no model description, create ouput failed");
        return FAILED;
    }

    output_ = aclmdlCreateDataset();
    if (output_ == nullptr) {
        ERROR_LOG("can't create dataset, create output failed");
        return FAILED;
    }

    size_t outputSize = aclmdlGetNumOutputs(modelDesc_);
    for (size_t i = 0; i < outputSize; ++i) {
        size_t modelOutputSize = aclmdlGetOutputSizeByIndex(modelDesc_, i);
    


        void *outputBuffer = nullptr;
        aclError ret = aclrtMalloc(&outputBuffer, modelOutputSize, ACL_MEM_MALLOC_NORMAL_ONLY);
        if (!Utils::validStatus("aclrtMalloc", ret)) {
            ERROR_LOG("Require output buffer size is %zu", modelOutputSize);
            return FAILED;
        }

        aclDataBuffer *outputData = aclCreateDataBuffer(outputBuffer, modelOutputSize);
        if (outputData == nullptr) {
            ERROR_LOG("can't create data buffer, create output failed");
            (void)aclrtFree(outputBuffer);
            return FAILED;
        }


        ret = aclmdlAddDatasetBuffer(output_, outputData);
        if (!Utils::validStatus("aclmdlAddDatasetBuffer", ret)) {
            ERROR_LOG("can't add data buffer for output");
            (void)aclrtFree(outputBuffer);
            (void)aclDestroyDataBuffer(outputData);
            return FAILED;
        }
    }

    INFO_LOG("create model output success");

    return SUCCESS;
}

void ModelProcess::DumpModelOutputResult()
{
    stringstream ss;
    size_t outputNum = aclmdlGetDatasetNumBuffers(output_);
    static int executeNum = 0;
    for (size_t i = 0; i < outputNum; ++i) {
        ss << "output" << ++executeNum << "_" << i << ".bin";
        string outputFileName = ss.str();
        FILE *outputFile = fopen(outputFileName.c_str(), "wb");
        if (outputFile != nullptr) {
            // get model output data
            aclDataBuffer *dataBuffer = aclmdlGetDatasetBuffer(output_, i);
            void *data = aclGetDataBufferAddr(dataBuffer);
            uint32_t len = aclGetDataBufferSizeV2(dataBuffer);

            void *outHostData = nullptr;
            aclError ret = ACL_SUCCESS;
            if (!g_isDevice) {
                ret = aclrtMallocHost(&outHostData, len);
                if (ret != ACL_SUCCESS) {
                    ERROR_LOG("aclrtMallocHost failed, malloc len[%u], errorCode[%d]",
                        len, static_cast<int32_t>(ret));
                    fclose(outputFile);
                    return;
                }

                // if app is running in host, need copy model output data from device to host
                ret = aclrtMemcpy(outHostData, len, data, len, ACL_MEMCPY_DEVICE_TO_HOST);
                if (ret != ACL_SUCCESS) {
                    ERROR_LOG("aclrtMemcpy failed, errorCode[%d]", static_cast<int32_t>(ret));
                    (void)aclrtFreeHost(outHostData);
                    fclose(outputFile);
                    return;
                }

                fwrite(outHostData, len, sizeof(char), outputFile);

                ret = aclrtFreeHost(outHostData);
                if (ret != ACL_SUCCESS) {
                    ERROR_LOG("aclrtFreeHost failed, errorCode[%d]", static_cast<int32_t>(ret));
                    fclose(outputFile);
                    return;
                }
            } else {
                // if app is running in host, write model output data into result file
                fwrite(data, len, sizeof(char), outputFile);
            }
            fclose(outputFile);
        } else {
            ERROR_LOG("create output file [%s] failed", outputFileName.c_str());
            return;
        }
    }

    INFO_LOG("dump data success");
    return;
}

//void ModelProcess::OutputModelResult()
std::vector<float*> ModelProcess::OutputModelResult()
{
    outputDataVector.clear();
    WARN_LOG("PRINTING THE RESULTS");
    for (size_t i = 0; i < aclmdlGetDatasetNumBuffers(output_); ++i) {
        // get model output data
        aclDataBuffer* dataBuffer = aclmdlGetDatasetBuffer(output_, i);
        void* data = aclGetDataBufferAddr(dataBuffer);
        uint32_t len = aclGetDataBufferSizeV2(dataBuffer);

        void *outHostData = nullptr;
        aclError ret = ACL_SUCCESS;
        float *outData = nullptr;
        outData = reinterpret_cast<float*>(data);
        outputDataVector.push_back(outData);
    //     if (!g_isDevice) {
    //         aclError ret = aclrtMallocHost(&outHostData, len);
    //         if (!Utils::validStatus("aclrtMallocHost", ret)) {
    //             ERROR_LOG("malloc len[%u]", len);
    //             // return;
    //         }

    //         // if app is running in host, need copy model output data from device to host
    //         ret = aclrtMemcpy(outHostData, len, data, len, ACL_MEMCPY_DEVICE_TO_HOST);
    //         if (!Utils::validStatus("aclrtMemcpy", ret)) {
    //             (void)aclrtFreeHost(outHostData);
    //             // return;
    //         }

    //         outData = reinterpret_cast<float*>(outHostData);
    //     } else {
    //         outData = reinterpret_cast<float*>(data);
    //     }

    //     outputDataVector.push_back(outData);
    //     for (unsigned int j = 0; j < len / sizeof(float); ++j) {
            
    //         if(j < 10)
    //         std::cout << "outData[" << j <<"]: "<< outData[j] << std::endl;
    //     }
    //     map<float, unsigned int, greater<float> > resultMap;
    //     for (unsigned int j = 0; j < len / sizeof(float); ++j) {
    //         resultMap[*outData] = j;
    //         outData++;
    //     }

    //     int cnt = 0;
    //     for (auto it = resultMap.begin(); it != resultMap.end(); ++it) {
    //         // print top 5
    //         if (++cnt > 5) {
    //             break;
    //         }

    //         INFO_LOG("top %d: index[%d] value[%lf]", cnt, it->second, it->first);
    //     }
    //     if (!g_isDevice) {
    //         ret = aclrtFreeHost(outHostData);
    //         if (!Utils::validStatus("aclrtFreeHost", ret)) {
    //             // return;
    //         }
    //     }
    }

    INFO_LOG("output data success");
    return outputDataVector;
}

void ModelProcess::DestroyOutput()
{
    if (output_ == nullptr) {
        return;
    }

    for (size_t i = 0; i < aclmdlGetDatasetNumBuffers(output_); ++i) {
        aclDataBuffer* dataBuffer = aclmdlGetDatasetBuffer(output_, i);
        void* data = aclGetDataBufferAddr(dataBuffer);
        (void)aclrtFree(data);
        (void)aclDestroyDataBuffer(dataBuffer);
    }

    (void)aclmdlDestroyDataset(output_);
    output_ = nullptr;
    INFO_LOG("destroy model output success");
}

Result ModelProcess::Execute()
{
    aclError ret = aclmdlExecute(modelId_, input_, output_);
    if (!Utils::validStatus("aclmdlExecute", ret)) {
        ERROR_LOG("modelId is %u", modelId_);
        return FAILED;
    }

    INFO_LOG("model execute success");
    return SUCCESS;
}

void ModelProcess::UnloadModel()
{
    if (!loadFlag_) {
        WARN_LOG("no model had been loaded, unload failed");
        return;
    }

    aclError ret = aclmdlUnload(modelId_);
    if (ret != ACL_SUCCESS) {
        ERROR_LOG("unload model failed, modelId is %u, errorCode is %d",
            modelId_, static_cast<int32_t>(ret));
    }

    if (modelDesc_ != nullptr) {
        (void)aclmdlDestroyDesc(modelDesc_);
        modelDesc_ = nullptr;
    }

    if (modelWorkPtr_ != nullptr) {
        (void)aclrtFree(modelWorkPtr_);
        modelWorkPtr_ = nullptr;
        modelWorkSize_ = 0;
    }

    if (modelWeightPtr_ != nullptr) {
        (void)aclrtFree(modelWeightPtr_);
        modelWeightPtr_ = nullptr;
        modelWeightSize_ = 0;
    }

    loadFlag_ = false;
    INFO_LOG("unload model success, modelId is %u", modelId_);
    modelId_ = 0;
}
